@@ -31,9 +31,10 @@ from typing import (
 
 import attr
 
-from hypothesis.errors import Frozen, InvalidArgument, StopTest
+from hypothesis.errors import ChoiceTooLarge, Frozen, InvalidArgument, StopTest
 from hypothesis.internal.cache import LRUCache
 from hypothesis.internal.compat import add_note, floor, int_from_bytes, int_to_bytes
+from hypothesis.internal.conjecture.choice import choice_from_index
 from hypothesis.internal.conjecture.floats import float_to_lex, lex_to_float
 from hypothesis.internal.conjecture.junkdrawer import (
     IntList,
@@ -58,6 +59,7 @@ from hypothesis.internal.floats import (
     sign_aware_lte,
 )
 from hypothesis.internal.intervalsets import IntervalSet
+from hypothesis.reporting import debug_report
 
 if TYPE_CHECKING:
     from typing import TypeAlias
@@ -231,7 +233,7 @@ class Example:
     often dramatically reduce our memory usage, so is worth it.
     """
 
-    __slots__ = ("owner", "index")
+    __slots__ = ("index", "owner")
 
     def __init__(self, owner: "Examples", index: int) -> None:
         self.owner = owner
@@ -356,8 +358,6 @@ class ExampleProperty:
                 self.block(self.block_count)
                 self.block_count += 1
             elif record == IR_NODE_RECORD:
-                data = self.examples.ir_nodes[self.ir_node_count]
-                self.ir_node(data)
                 self.ir_node_count += 1
             elif record >= START_EXAMPLE_RECORD:
                 self.__push(record - START_EXAMPLE_RECORD)
@@ -399,9 +399,6 @@ class ExampleProperty:
         index of the example and ``discarded`` being ``True`` if ``stop_example``
         was called with ``discard=True``."""
 
-    def ir_node(self, node: "IRNode") -> None:
-        """Called when an ir node is drawn."""
-
     def finish(self) -> Any:
         return self.result
 
@@ -457,23 +454,8 @@ class ExampleRecord:
     def freeze(self) -> None:
         self.__index_of_labels = None
 
-    def record_ir_draw(
-        self,
-        ir_type: IRTypeName,
-        value: IRType,
-        *,
-        kwargs: IRKWargsType,
-        was_forced: bool,
-    ) -> None:
+    def record_ir_draw(self) -> None:
         self.trail.append(IR_NODE_RECORD)
-        node = IRNode(
-            ir_type=ir_type,
-            value=value,
-            kwargs=kwargs,
-            was_forced=was_forced,
-            index=len(self.ir_nodes),
-        )
-        self.ir_nodes.append(node)
 
     def start_example(self, label: int) -> None:
         assert self.__index_of_labels is not None
@@ -507,7 +489,6 @@ class Examples:
 
     def __init__(self, record: ExampleRecord, blocks: "Blocks") -> None:
         self.trail = record.trail
-        self.ir_nodes = record.ir_nodes
         self.labels = record.labels
         self.__length = self.trail.count(
             STOP_EXAMPLE_DISCARD_RECORD
@@ -617,18 +598,6 @@ class Examples:
 
     depths: IntList = calculated_example_property(_depths)
 
-    class _ir_tree_nodes(ExampleProperty):
-        def begin(self) -> None:
-            self.result = []
-
-        def ir_node(self, ir_node: "IRNode") -> None:
-            self.result.append(ir_node)
-
-        def finish(self):
-            return tuple(self.result)
-
-    ir_tree_nodes: "tuple[IRNode, ...]" = calculated_example_property(_ir_tree_nodes)
-
     class _label_indices(ExampleProperty):
         def start_example(self, i: int, label_index: int) -> None:
             self.result[i] = label_index
@@ -745,7 +714,7 @@ class Blocks:
     be preferred to using the Block objects directly, as it will not
     have to allocate the actual object."""
 
-    __slots__ = ("endpoints", "owner", "__blocks", "__count", "__sparse")
+    __slots__ = ("__blocks", "__count", "__sparse", "endpoints", "owner")
     owner: "Union[ConjectureData, ConjectureResult, None]"
     __blocks: Union[dict[int, Block], list[Optional[Block]]]
 
@@ -997,18 +966,10 @@ class IRNode:
         if self.was_forced:
             return True
 
-        if self.ir_type == "integer":
-            shrink_towards = self.kwargs["shrink_towards"]
-            min_value = self.kwargs["min_value"]
-            max_value = self.kwargs["max_value"]
-
-            if min_value is not None:
-                shrink_towards = max(min_value, shrink_towards)
-            if max_value is not None:
-                shrink_towards = min(max_value, shrink_towards)
-
-            return self.value == shrink_towards
-        if self.ir_type == "float":
+        if self.ir_type != "float":
+            zero_value = choice_from_index(0, self.ir_type, self.kwargs)
+            return ir_value_equal(self.ir_type, self.value, zero_value)
+        else:
             min_value = self.kwargs["min_value"]
             max_value = self.kwargs["max_value"]
             shrink_towards = 0
@@ -1032,17 +993,6 @@ class IRNode:
             # It would be good to compute this correctly in the future, but it's
             # also not incorrect to be conservative here.
             return False
-        if self.ir_type == "boolean":
-            return self.value is False
-        if self.ir_type == "string":
-            # smallest size and contains only the smallest-in-shrink-order character.
-            minimal_char = self.kwargs["intervals"].char_in_shrink_order(0)
-            return self.value == (minimal_char * self.kwargs["min_size"])
-        if self.ir_type == "bytes":
-            # smallest size and all-zero value.
-            return len(self.value) == self.kwargs["min_size"] and not any(self.value)
-
-        raise NotImplementedError(f"unhandled ir_type {self.ir_type}")
 
     def __eq__(self, other):
         if not isinstance(other, IRNode):
@@ -1071,6 +1021,15 @@ class IRNode:
         return f"{self.ir_type} {self.value!r}{forced_marker} {self.kwargs!r}"
 
 
+@attr.s(slots=True)
+class NodeTemplate:
+    type: Literal["simplest"] = attr.ib()
+    size: int = attr.ib()
+
+    def __attrs_post_init__(self) -> None:
+        assert self.size > 0
+
+
 def ir_value_permitted(value, ir_type, kwargs):
     if ir_type == "integer":
         min_value = kwargs["min_value"]
@@ -1118,8 +1077,11 @@ def ir_size(ir: Iterable[IRType]) -> int:
     return len(ir_to_bytes(ir))
 
 
-def ir_size_nodes(nodes: Iterable[IRNode]) -> int:
-    return ir_size([n.value for n in nodes])
+def ir_size_nodes(nodes: Iterable[Union[IRNode, NodeTemplate]]) -> int:
+    size = 0
+    for node in nodes:
+        size += node.size if isinstance(node, NodeTemplate) else ir_size([node.value])
+    return size
 
 
 def ir_value_key(ir_type, v):
@@ -1173,6 +1135,7 @@ class ConjectureResult:
     # If we consider blocks or examples in equality checks, multiple semantically equal
     # results get stored in e.g. the pareto front.
     blocks: Blocks = attr.ib(eq=False)
+    ir_nodes: tuple[IRNode, ...] = attr.ib(eq=False, repr=False)
     output: str = attr.ib()
     extra_information: Optional[ExtraInformation] = attr.ib()
     has_discards: bool = attr.ib()
@@ -1193,6 +1156,10 @@ class ConjectureResult:
     def as_result(self) -> "ConjectureResult":
         return self
 
+    @property
+    def choices(self) -> tuple[IRType, ...]:
+        return tuple(node.value for node in self.ir_nodes)
+
 
 # Masks for masking off the first byte of an n-bit buffer.
 # The appropriate mask is stored at position n % 8.
@@ -1348,6 +1315,27 @@ class PrimitiveProvider(abc.ABC):
     ) -> bytes:
         raise NotImplementedError
 
+    def span_start(self, label: int, /) -> None:  # noqa: B027  # non-abstract noop
+        """Marks the beginning of a semantically meaningful span.
+
+        Providers can optionally track this data to learn which sub-sequences
+        of draws correspond to a higher-level object, recovering the parse tree.
+        `label` is an opaque integer, which will be shared by all spans drawn
+        from a particular strategy.
+
+        This method is called from ConjectureData.start_example().
+        """
+
+    def span_end(self, discard: bool, /) -> None:  # noqa: B027  # non-abstract noop
+        """Marks the end of a semantically meaningful span.
+
+        `discard` is True when the draw was filtered out or otherwise marked as
+        unlikely to contribute to the input data as seen by the user's test.
+        Note however that side effects can make this determination unsound.
+
+        This method is called from ConjectureData.stop_example().
+        """
+
 
 class HypothesisProvider(PrimitiveProvider):
     lifetime = "test_case"
@@ -1562,8 +1550,7 @@ class HypothesisProvider(PrimitiveProvider):
         (
             sampler,
             forced_sign_bit,
-            neg_clamper,
-            pos_clamper,
+            clamper,
             nasty_floats,
         ) = self._draw_float_init_logic(
             min_value=min_value,
@@ -1594,12 +1581,8 @@ class HypothesisProvider(PrimitiveProvider):
                 )
                 if allow_nan and math.isnan(result):
                     clamped = result
-                elif math.copysign(1.0, result) == -1:
-                    assert neg_clamper is not None
-                    clamped = -neg_clamper(-result)
                 else:
-                    assert pos_clamper is not None
-                    clamped = pos_clamper(result)
+                    clamped = clamper(result)
                 if clamped != result and not (math.isnan(result) and allow_nan):
                     self._draw_float(forced=clamped, fake_forced=fake_forced)
                     result = clamped
@@ -1881,8 +1864,7 @@ class HypothesisProvider(PrimitiveProvider):
     ) -> tuple[
         Optional[Sampler],
         Optional[Literal[0, 1]],
-        Optional[Callable[[float], float]],
-        Optional[Callable[[float], float]],
+        Callable[[float], float],
         list[float],
     ]:
         """
@@ -1919,8 +1901,7 @@ class HypothesisProvider(PrimitiveProvider):
     ) -> tuple[
         Optional[Sampler],
         Optional[Literal[0, 1]],
-        Optional[Callable[[float], float]],
-        Optional[Callable[[float], float]],
+        Callable[[float], float],
         list[float],
     ]:
         if smallest_nonzero_magnitude == 0.0:  # pragma: no cover
@@ -1952,23 +1933,17 @@ class HypothesisProvider(PrimitiveProvider):
         weights = [0.2 * len(nasty_floats)] + [0.8] * len(nasty_floats)
         sampler = Sampler(weights, observe=False) if nasty_floats else None
 
-        pos_clamper = neg_clamper = None
-        if sign_aware_lte(0.0, max_value):
-            pos_min = max(min_value, smallest_nonzero_magnitude)
-            allow_zero = sign_aware_lte(min_value, 0.0)
-            pos_clamper = make_float_clamper(pos_min, max_value, allow_zero=allow_zero)
-        if sign_aware_lte(min_value, -0.0):
-            neg_max = min(max_value, -smallest_nonzero_magnitude)
-            allow_zero = sign_aware_lte(-0.0, max_value)
-            neg_clamper = make_float_clamper(
-                -neg_max, -min_value, allow_zero=allow_zero
-            )
-
         forced_sign_bit: Optional[Literal[0, 1]] = None
-        if (pos_clamper is None) != (neg_clamper is None):
-            forced_sign_bit = 1 if neg_clamper else 0
+        if sign_aware_lte(min_value, -0.0) != sign_aware_lte(0.0, max_value):
+            forced_sign_bit = 1 if sign_aware_lte(min_value, -0.0) else 0
 
-        return (sampler, forced_sign_bit, neg_clamper, pos_clamper, nasty_floats)
+        clamper = make_float_clamper(
+            min_value,
+            max_value,
+            smallest_nonzero_magnitude=smallest_nonzero_magnitude,
+            allow_nan=allow_nan,
+        )
+        return (sampler, forced_sign_bit, clamper, nasty_floats)
 
 
 # The set of available `PrimitiveProvider`s, by name.  Other libraries, such as
@@ -1999,7 +1974,7 @@ class ConjectureData:
     @classmethod
     def for_ir_tree(
         cls,
-        ir_tree_prefix: Sequence[IRNode],
+        ir_tree_prefix: Sequence[Union[IRNode, NodeTemplate]],
         *,
         observer: Optional[DataObserver] = None,
         provider: Union[type, PrimitiveProvider] = HypothesisProvider,
@@ -2028,7 +2003,7 @@ class ConjectureData:
         random: Optional[Random],
         observer: Optional[DataObserver] = None,
         provider: Union[type, PrimitiveProvider] = HypothesisProvider,
-        ir_tree_prefix: Optional[Sequence[IRNode]] = None,
+        ir_tree_prefix: Optional[Sequence[Union[IRNode, NodeTemplate]]] = None,
         max_length_ir: Optional[int] = None,
     ) -> None:
         from hypothesis.internal.conjecture.engine import BUFFER_SIZE_IR
@@ -2106,7 +2081,8 @@ class ConjectureData:
 
         self.extra_information = ExtraInformation()
 
-        self.ir_tree_nodes = ir_tree_prefix
+        self.ir_prefix = ir_tree_prefix
+        self.ir_nodes: tuple[IRNode, ...] = ()
         self.misaligned_at: Optional[MisalignedAt] = None
         self.start_example(TOP_LABEL)
 
@@ -2117,6 +2093,10 @@ class ConjectureData:
             ", frozen" if self.frozen else "",
         )
 
+    @property
+    def choices(self) -> tuple[IRType, ...]:
+        return tuple(node.value for node in self.ir_nodes)
+
     # A bit of explanation of the `observe` and `fake_forced` arguments in our
     # draw_* functions.
     #
@@ -2147,10 +2127,11 @@ class ConjectureData:
         # end of the function, but avoids trying to use a null self.random when
         # drawing past the node of a ConjectureData.for_ir_tree data.
         if self.length_ir == self.max_length_ir:
+            debug_report(f"overrun because hit {self.max_length_ir=}")
             self.mark_overrun()
 
-        if self.ir_tree_nodes is not None and observe:
-            if self.index_ir < len(self.ir_tree_nodes):
+        if self.ir_prefix is not None and observe:
+            if self.index_ir < len(self.ir_prefix):
                 node_value = self._pop_ir_tree_node(ir_type, kwargs, forced=forced)
             else:
                 try:
@@ -2158,6 +2139,7 @@ class ConjectureData:
                         ir_type, kwargs, forced=forced, random=self.__random
                     )
                 except StopTest:
+                    debug_report("overrun because ir_to_buffer overran")
                     self.mark_overrun()
 
             if forced is None:
@@ -2173,17 +2155,24 @@ class ConjectureData:
             getattr(self.observer, f"draw_{ir_type}")(
                 value, kwargs=kwargs, was_forced=was_forced
             )
-            self.__example_record.record_ir_draw(
-                ir_type,
-                value,
+            size = ir_size([value])
+            if self.length_ir + size > self.max_length_ir:
+                debug_report(
+                    f"overrun because {self.length_ir=} + {size=} > {self.max_length_ir=}"
+                )
+                self.mark_overrun()
+
+            node = IRNode(
+                ir_type=ir_type,
+                value=value,
                 kwargs=kwargs,
                 was_forced=was_forced,
+                index=len(self.ir_nodes),
             )
-            self.index_ir += 1
-            self.length_ir += ir_size([value])
+            self.__example_record.record_ir_draw()
+            self.ir_nodes += (node,)
+            self.length_ir += size
 
-        if self.length_ir > self.max_length_ir:
-            self.mark_overrun()
         return value
 
     def draw_integer(
@@ -2362,11 +2351,30 @@ class ConjectureData:
     ) -> IRType:
         from hypothesis.internal.conjecture.engine import BUFFER_SIZE
 
-        assert self.ir_tree_nodes is not None
+        assert self.ir_prefix is not None
         # checked in _draw
-        assert self.index_ir < len(self.ir_tree_nodes)
+        assert self.index_ir < len(self.ir_prefix)
+
+        node = self.ir_prefix[self.index_ir]
+        if isinstance(node, NodeTemplate):
+            assert node.size >= 0
+            # node templates have to be at the end for now, since it's not immediately
+            # apparent how to handle overruning a node template while generating a single
+            # node if the alternative is not "the entire data is an overrun".
+            assert self.index_ir == len(self.ir_prefix) - 1
+            if node.type == "simplest":
+                try:
+                    value = choice_from_index(0, ir_type, kwargs)
+                except ChoiceTooLarge:
+                    self.mark_overrun()
+            else:
+                raise NotImplementedError
+
+            node.size -= ir_size([value])
+            if node.size < 0:
+                self.mark_overrun()
+            return value
 
-        node = self.ir_tree_nodes[self.index_ir]
         value = node.value
         # If we're trying to:
         # * draw a different ir type at the same location
@@ -2407,6 +2415,7 @@ class ConjectureData:
                 #   buffer_to_ir(ir_type, kwargs, buffer=bytes(BUFFER_SIZE))
                 self.mark_overrun()
 
+        self.index_ir += 1
         return value
 
     def as_result(self) -> Union[ConjectureResult, _Overrun]:
@@ -2422,6 +2431,7 @@ class ConjectureData:
                 interesting_origin=self.interesting_origin,
                 buffer=self.buffer,
                 examples=self.examples,
+                ir_nodes=self.ir_nodes,
                 blocks=self.blocks,
                 output=self.output,
                 extra_information=(
@@ -2457,6 +2467,9 @@ class ConjectureData:
         label: Optional[int] = None,
         observe_as: Optional[str] = None,
     ) -> "Ex":
+        from hypothesis.internal.observability import TESTCASE_CALLBACKS
+        from hypothesis.strategies._internal.utils import to_jsonable
+
         if self.is_find and not strategy.supports_find:
             raise InvalidArgument(
                 f"Cannot use strategy {strategy!r} within a call to find "
@@ -2493,19 +2506,26 @@ class ConjectureData:
             try:
                 strategy.validate()
                 try:
-                    return strategy.do_draw(self)
+                    v = strategy.do_draw(self)
                 finally:
                     # Subtract the time spent in GC to avoid overcounting, as it is
                     # accounted for at the overall example level.
                     in_gctime = gc_cumulative_time() - gc_start_time
                     self.draw_times[key] = time.perf_counter() - start_time - in_gctime
             except Exception as err:
-                add_note(err, f"while generating {key[9:]!r} from {strategy!r}")
+                add_note(
+                    err,
+                    f"while generating {key.removeprefix('generate:')!r} from {strategy!r}",
+                )
                 raise
+            if TESTCASE_CALLBACKS:
+                self._observability_args[key] = to_jsonable(v)
+            return v
         finally:
             self.stop_example()
 
     def start_example(self, label: int) -> None:
+        self.provider.span_start(label)
         self.__assert_not_frozen("start_example")
         self.depth += 1
         # Logically it would make sense for this to just be
@@ -2520,6 +2540,7 @@ class ConjectureData:
         self.labels_for_structure_stack.append({label})
 
     def stop_example(self, *, discard: bool = False) -> None:
+        self.provider.span_end(discard)
         if self.frozen:
             return
         if discard:
