@@ -269,12 +269,6 @@ class LazySequenceCopy:
         return i
 
 
-def clamp(lower: float, value: float, upper: float) -> float:
-    """Given a value and lower/upper bounds, 'clamp' the value so that
-    it satisfies lower <= value <= upper."""
-    return max(lower, min(value, upper))
-
-
 def swap(ls: LazySequenceCopy, i: int, j: int) -> None:
     """Swap the elements ls[i], ls[j]."""
     if i == j:
